import argparse
import asyncio
import codecs
import json
import os
import sys

from . import Client, Format, Request, ResponseFlags


def parse_arguments():
    proxy_required = False

    try:
        if int(os.environ.get("RPC_PROXY", 0)):
            proxy_required = True
    except:
        print("Invalid environment configuration.")
        sys.exit(1)

    # fmt: off
    parser = argparse.ArgumentParser(
        description="RPC client",
        usage="rpc example.com:2000 --token token --name function",
    )
    parser.add_argument(
        "target",
        type=str,
        help="Server address.",
    )
    parser.add_argument(
        "--proxy",
        required=proxy_required,
        type=str,
        help="Socks5 proxy addres with optional credentials.",
    )
    parser.add_argument(
        "--tls-ca",
        required=proxy_required,
        type=str,
        help="TLS ca certificate path.",
    )
    parser.add_argument(
        "--tls-hostname",
        required=proxy_required,
        type=str,
        help="Hostname.",
    )
    parser.add_argument(
        "--token",
        type=str,
        help="Token.",
    )
    parser.add_argument(
        "--name",
        type=str,
        help="Name.",
    )
    parser.add_argument(
        "--version",
        type=int,
        help="Version.",
    )
    parser.add_argument(
        "--format",
        type=str,
        choices=("binary", "base64", "pickle", "json"),
        help="Format.",
    )
    parser.add_argument(
        "--data-command-line",
        type=str,
        help="Data (from command line).",
    )
    parser.add_argument(
        "--data-file",
        type=str,
        help="Data (from file, overrides --data-command-line value).",
    )
    parser.add_argument(
        "--documentation",
        action="store_true",
        help="Documentation.",
    )
    parser.add_argument(
        "--deferred",
        action="store_true",
        help="Deferred.",
    )
    parser.add_argument(
        "--timeout",
        type=float,
        help="Request time limit.",
    )
    parser.add_argument(
        "--status",
        action="store_true",
        help="Print response status.",
    )
    # fmt: on
    return parser.parse_args()


def parse_parameters():
    arguments = parse_arguments()

    if ":" in arguments.target:
        target = arguments.target.split(":")

        if len(target) != 2:
            print("Invalid target format.")
            sys.exit(1)

        host = target[0]
        port = int(target[1])
    else:
        host = arguments.target
        port = 2000

    proxy = None
    if arguments.proxy:
        if "@" in arguments.proxy:
            proxy_pair = arguments.proxy.split("@")

            if len(proxy_pair) != 2:
                print("Invalid proxy format.")
                sys.exit(1)

            proxy_credentials = proxy_pair[0].split(":")
            proxy_target = proxy_pair[1].split(":")
        else:
            proxy_credentials = None
            proxy_target = arguments.proxy.split(":")

        if proxy_credentials and len(proxy_credentials) != 2:
            print("Invalid proxy format.")
            sys.exit(1)

        if len(proxy_target) != 2:
            print("Invalid proxy format.")
            sys.exit(1)

        try:
            proxy = {
                "host": proxy_target[0],
                "port": int(proxy_target[1]),
            }
        except:
            print("Invalid proxy format.")
            sys.exit(1)

        if proxy_credentials:
            proxy.update({"user": proxy_credentials[0], "password": proxy_credentials[1]})

    tls = None
    if arguments.tls_ca and arguments.tls_hostname:
        tls = {
            "ca": arguments.tls_ca,
            "hostname": arguments.tls_hostname,
            "ciphers": [
                "ECDHE-RSA-AES128-GCM-SHA256",
                "ECDHE-RSA-AES256-GCM-SHA384",
                "ECDHE-RSA-AES128-SHA",
                "ECDHE-RSA-AES256-SHA",
                "ECDHE-RSA-AES128-SHA256",
                "ECDHE-RSA-AES256-SHA384",
                "DHE-RSA-AES128-GCM-SHA256",
                "DHE-RSA-AES256-GCM-SHA384",
                "DHE-RSA-AES128-SHA",
                "DHE-RSA-AES256-SHA",
                "DHE-RSA-AES128-SHA256",
                "DHE-RSA-AES256-SHA256",
            ],
        }

    token = None
    if arguments.token:
        token = arguments.token

    version = None
    if arguments.version:
        version = arguments.version

    name = None
    if arguments.name:
        name = arguments.name

    format = "pickle"
    if arguments.format:
        format = arguments.format

    if format not in ("binary", "base64", "pickle", "json"):
        print("Invalid format format.")
        sys.exit(1)

    format = Format.from_string(format)

    data = None
    if arguments.data_command_line:
        data = arguments.data_command_line

    if arguments.data_file:
        if format == Format.BINARY:
            mode = "rb"

        if format == Format.BASE64:
            mode = "rb"

        if format == Format.PICKLE:
            mode = "rt"

        if format == Format.JSON:
            mode = "rt"

        try:
            with open(arguments.data_file, mode) as handle:
                data = handle.read()
        except:
            print("Error reading file.")
            sys.exit(1)

    if data:
        if format == Format.BINARY:
            data = codecs.decode(data, "unicode_escape").encode()

        if format == Format.BASE64:
            data = codecs.decode(data, "unicode_escape").encode()

        if format == Format.PICKLE:
            data = eval(data)

        if format == Format.JSON:
            data = json.loads(data)

    documentation = False
    if arguments.documentation:
        documentation = True

    deferred = False
    if arguments.deferred:
        deferred = True

    timeout = 10
    if arguments.timeout:
        timeout = arguments.timeout

    status = False
    if arguments.status:
        status = True

    return host, port, proxy, tls, token, version, name, format, data, documentation, deferred, timeout, status


def print_result(response, status):
    output = False

    if response.flags & ResponseFlags.DATA:
        output = True

        if response.format == Format.BINARY:
            print(response.data)

        if response.format == Format.BASE64:
            print(response.data)

        if response.format == Format.PICKLE:
            print(json.dumps(response.data, indent=4))

        if response.format == Format.JSON:
            print(json.dumps(response.data, indent=4))

    if response.flags & ResponseFlags.DOCUMENTATION:
        output = True
        print(response.documentation)

    if status:
        if output:
            print("--------------------------------------------------------------------------------")

        print(f"Status: {response.status}.")

    exit(response.status)


async def main():
    (
        host,
        port,
        proxy,
        tls,
        token,
        version,
        name,
        format,
        data,
        documentation,
        deferred,
        timeout,
        status,
    ) = parse_parameters()

    client = Client(host=host, port=port, proxy=proxy, tls=tls, token=token)
    request = Request(
        version=version, name=name, format=format, data=data, documentation=documentation, deferred=deferred
    )

    try:
        response = await client.send_request(request, timeout=timeout)
    except Exception as e:
        print(f"Connection error `{e}`.")
        sys.exit(1)

    print_result(response, status)


try:
    asyncio.run(main())
except:
    pass
