import asyncio
import logging

import aiosocks

from .context import set_client
from .errors import ProtocolError
from .format import Format
from .request import Request
from .response import Response, Status
from .wrappers import ReaderWrapper, WriterWrapper

logger = logging.getLogger(__name__)


class Client:
    def __init__(
        self,
        *,
        name="default",
        host="127.0.0.1",
        port=3000,
        proxy=None,
        tls=None,
        token=None,
        data_length_max=65535,
        documentation_length_max=65535,
        timeout=10,
    ):
        self.name = name
        self.host = host
        self.port = port
        self.proxy = proxy
        self.tls = tls
        self.token = token
        self.data_length_max = data_length_max
        self.documentation_length_max = documentation_length_max
        self.timeout = timeout

        set_client(self)

    async def send(
        self,
        name,
        version=None,
        timeout=10,
        retries=None,
        status=Status.SUCCESS,
        format=Format.PICKLE,
        **kwargs,
    ):
        attempts = 0
        sleep = None

        if retries is not None:
            sleep = retries.delay_min

        while True:
            try:
                request = Request(
                    version=version, name=name, format=format, data=kwargs
                )
                response = await self.send_request(request, timeout=timeout)

                if response.status != status:
                    if response.status == Status.FAIL:
                        retriable = True
                    else:
                        retriable = False

                    raise ProtocolError(
                        "Invalid response status.", retriable=retriable
                    )

                return response.data
            except ProtocolError as exception:
                if not exception.retriable:
                    raise exception

                if retries is None:
                    raise exception

                attempts += 1

                if attempts > retries.attempts:
                    raise exception

                await asyncio.sleep(sleep)
                sleep = retries.next_sleep(sleep)

    async def send_request(self, request, *, timeout=None):
        request.token = self.token

        if timeout is None:
            timeout = self.timeout

        try:
            return await asyncio.wait_for(self._send(request), timeout=timeout)
        except asyncio.TimeoutError as exception:
            raise ProtocolError("Send time is out.", retriable=True)
        except ConnectionError as exception:
            raise ProtocolError("Connection error.", retriable=True)
        except Exception as exception:
            context = {
                "exception": exception,
            }
            logger.exception(
                f"[Client] Exception while executing request. Context `{context}`."
            )
            raise ProtocolError(exception)

    async def _send(self, request):
        # Prepare connection.
        if self.proxy:
            connection_parameters = {
                "dst": (self.host, self.port),
                "proxy": aiosocks.Socks5Addr(
                    self.proxy["host"], self.proxy["port"]
                ),
                "proxy_auth": None,
                "remote_resolve": True,
            }

            if "user" in self.proxy:
                connection_parameters["proxy_auth"] = aiosocks.Socks5Auth(
                    self.proxy["user"], self.proxy["password"]
                )

            # if self.tls:
            #     connection_parameters["ssl"] = create_tls_context(server=False, configuration=self.tls)
            #     connection_parameters["server_hostname"] = self.tls["hostname"]

            reader, writer = await aiosocks.open_connection(
                **connection_parameters
            )
        else:
            connection_parameters = {
                "host": self.host,
                "port": self.port,
            }

            # if self.tls:
            #     connection_parameters["ssl"] = create_tls_context(server=False, configuration=self.tls)
            #     connection_parameters["server_hostname"] = self.tls["hostname"]

            reader, writer = await asyncio.open_connection(
                **connection_parameters
            )

        reader = ReaderWrapper(reader)
        writer = WriterWrapper(writer)

        # Send request.
        await request.write(writer)

        # Read response.
        response = await Response.read(
            reader,
            data_length_max=self.data_length_max,
            documentation_length_max=self.documentation_length_max,
        )

        # Close connection.
        writer.close()

        # Return response.
        return response
