import inspect
from functools import partial

from .context import add_handler, get_client
from .format import Format
from .handler import Handler
from .response import Status


def remote(
    *args,
    version=None,
    timeout=10,
    retries=None,
    status=Status.SUCCESS,
    format=Format.PICKLE,
):
    def _wrapper(function):
        name = _get_client_name(function)

        async def __wrapper(**kwargs):
            return await get_client().send(name, version, timeout, retries, status, format, **kwargs)

        return __wrapper

    if len(args):
        return _wrapper(args[0])

    return _wrapper


def endpoint(
    *args,
    version=0,
    permission=None,
    format=Format.PICKLE,
    schema_in=None,
    schema_out=None,
    data_length_max=65535,
    timeout=10,
    allow_deferred=False,
):
    def _wrapper(function):
        async def __wrapper(*args, **kwargs):
            return await function(*args, **kwargs)

        signature = inspect.signature(function)
        name = _get_server_name(function)
        parameters = list(signature.parameters)

        if "user" in parameters:
            parameters.remove("user")

        parameters = ", ".join(parameters)
        documentation = ""
        docstring = function.__doc__

        if permission:
            documentation += f"@{permission}\n"

        if parameters:
            documentation += f"{name} ({parameters})"
        else:
            documentation += f"{name}"

        if docstring:
            documentation += f"{docstring}"
        else:
            documentation += "\n"

        handler = Handler(
            version=version,
            permission=permission,
            function=__wrapper,
            name=name,
            parameters=signature.parameters,
            documentation=documentation,
            format=format,
            schema_in=schema_in,
            schema_out=schema_out,
            data_length_max=data_length_max,
            timeout=timeout,
            allow_deferred=allow_deferred,
        )
        add_handler(version, name, handler)
        return __wrapper

    if len(args):
        return _wrapper(args[0])

    return _wrapper


endpoint_deferred = partial(endpoint, allow_deferred=True)


def _get_client_name(function):
    try:
        path = function.__module__.split(".")
        index = path.index("rpc_client")
        path = path[index + 1 :]

        assert len(path) > 0, "Invalid location for remote."

        return ".".join(path)
    except:
        raise Exception("Remotes should be in rpc_client directory.")


def _get_server_name(function):
    try:
        path = function.__module__.split(".")
        index = path.index("rpc_server")
        path = path[index + 1 :]

        assert len(path) > 0, "Invalid location for endpoint."

        return ".".join(path)
    except:
        raise Exception("Endpoints should be in rpc_server directory.")
