import random
from dataclasses import dataclass

from .errors import ProtocolError
from .format import Format


class RequestFlags:
    TOKEN = 0x01
    VERSION = 0x02
    NAME = 0x04
    DATA = 0x08
    DOCUMENTATION = 0x10
    DEFERRED = 0x20


@dataclass
class Retry:
    attempts: int = 3
    delay_min: float = 1.0
    delay_max: float = 10.0
    factor: float = 2.0
    jitter: float = 0.5

    def next_sleep(self, previous_sleep):
        sleep = min(previous_sleep * self.factor, self.delay_max)
        return sleep + random.uniform(0, sleep * self.jitter)


@dataclass
class Request:
    flags: int
    token: str
    version: int
    name: str
    format: int
    data: object

    def __init__(self, *, version=0, name=None, format=None, data=None, documentation=False, deferred=False):
        self.flags = 0
        self.token = None
        self.version = version
        self.name = name
        self.format = format
        self.data = data
        self.documentation = documentation
        self.deferred = deferred

    @staticmethod
    async def read(reader, *, token_length_max, name_length_max, data_length_max):
        request = Request()
        request.flags = await reader.read_byte()

        if request.flags & RequestFlags.TOKEN:
            request.token = await reader.read_bytes(True, token_length_max)

        if request.flags & RequestFlags.VERSION:
            request.version = await reader.read_byte()
        else:
            request.version = 0

        if request.flags & RequestFlags.NAME:
            request.name = await reader.read_bytes(True, name_length_max)

        if request.name is None:
            return request

        if request.version not in data_length_max:
            raise ProtocolError("Function unknown.")

        if request.name not in data_length_max[request.version]:
            raise ProtocolError("Function unknown.")

        if request.flags & RequestFlags.DATA:
            request.format = await reader.read_byte()

            if request.format not in (
                Format.BINARY,
                Format.BASE64,
                Format.PICKLE,
                Format.JSON,
            ):
                raise ProtocolError("Invalid request format.")

            request.data = await reader.read_bytes(request.format, data_length_max[request.version][request.name])

        if request.flags & RequestFlags.DOCUMENTATION:
            request.documentation = True

        if request.flags & RequestFlags.DEFERRED:
            request.deferred = True

        return request

    async def write(self, writer):
        self._build_flags()
        writer.write_byte(self.flags)

        if self.token:
            writer.write_bytes(True, self.token)

        if self.version:
            writer.write_byte(self.version)

        if self.name:
            writer.write_bytes(True, self.name)

        if self.data:
            writer.write_byte(self.format)
            writer.write_bytes(self.format, self.data)

        await writer.drain()

    def _build_flags(self):
        self.flags = 0

        if self.token:
            self.flags |= RequestFlags.TOKEN

        if self.version:
            self.flags |= RequestFlags.VERSION

        if self.name:
            self.flags |= RequestFlags.NAME

        if self.data:
            assert format, "Format not set."
            self.flags |= RequestFlags.DATA

        if self.documentation:
            self.flags |= RequestFlags.DOCUMENTATION

        if self.deferred:
            self.flags |= RequestFlags.DEFERRED
