from dataclasses import dataclass

from .errors import ProtocolError
from .format import Format


class ResponseFlags:
    DATA = 0x01
    DOCUMENTATION = 0x02


class Status:
    SUCCESS = 0
    FAIL = 1
    FUNCTION_NOT_FOUND = 2
    FUNCTION_INVALID_SIGNATURE = 3
    AUTHENTICATION_MISSING_TOKEN = 10
    AUTHENTICATION_USER_NOT_FOUND = 11
    AUTHENTICATION_PERMISSION_DENIED = 12
    AUTHENTICATION_OUT_OF_CAPACITY = 13
    MARSHALLING_IN_FAILED = 20
    MARSHALLING_OUT_FAILED = 21


@dataclass
class Response:
    flags: int
    status: int
    format: int
    data: object
    documentation: str

    def __init__(self, *, status, format=None, data=None, documentation=None):
        self.flags = 0
        self.status = status
        self.format = format
        self.data = data
        self.documentation = documentation

    @staticmethod
    async def read(reader, *, data_length_max, documentation_length_max):
        response = Response(status=None)

        response.flags = await reader.read_byte()
        response.status = await reader.read_byte()

        if response.flags & ResponseFlags.DATA:
            response.format = await reader.read_byte()

            if response.format not in (Format.BINARY, Format.BASE64, Format.PICKLE, Format.JSON):
                raise ProtocolError("Invalid response format.")

            response.data = await reader.read_bytes(response.format, data_length_max)

        if response.flags & ResponseFlags.DOCUMENTATION:
            response.documentation = await reader.read_bytes(True, documentation_length_max)

        return response

    async def write(self, writer):
        self._build_flags()

        writer.write_byte(self.flags)
        writer.write_byte(self.status)

        if self.flags & ResponseFlags.DATA:
            writer.write_byte(self.format)
            writer.write_bytes(self.format, self.data)

        if self.flags & ResponseFlags.DOCUMENTATION:
            writer.write_bytes(True, self.documentation)

        await writer.drain()

    def _build_flags(self):
        self.flags = 0

        if self.data:
            assert format, "Format not set."
            self.flags |= ResponseFlags.DATA

        if self.documentation:
            self.flags |= ResponseFlags.DOCUMENTATION
