import asyncio
import logging
import time
from collections import defaultdict

import marshmallow

from .authentication import DummyAuthenticationBackend
from .context import get_handlers, set_server, set_token, set_user
from .deferred import InMemoryDeferredBackend
from .format import Format
from .handler import Handler
from .helpers import asynchronous_context, create_tls_context
from .request import Request
from .response import Response, Status
from .wrappers import ReaderWrapper, WriterWrapper

logger = logging.getLogger(__name__)


class Server:
    def __init__(
        self,
        *,
        name="default",
        host="127.0.0.1",
        port=3000,
        tls=None,
        response_format=Format.PICKLE,
        name_length_max=100,
        token_length_max=100,
        data_length_max=65535,
        timeout=10,
        authentication_backend=DummyAuthenticationBackend(),
        deferred_backend=InMemoryDeferredBackend(),
        deferred=False,
    ):
        self.name = name
        self.host = host
        self.port = port
        self.tls = tls
        self.response_format = response_format
        self.name_length_max = name_length_max
        self.token_length_max = token_length_max
        self.data_length_max = data_length_max
        self.timeout = timeout
        self.server = None

        self.authentication_backend = authentication_backend
        self.deferred_backend = deferred_backend
        self.deferred = deferred

        self.handlers = get_handlers()
        self.add_default_handlers()

        set_server(self)

    def add_handler(self, version, name, handler):
        self.handlers[version][name] = handler

    def add_default_handlers(self):
        self.add_handler(
            0,
            "ping",
            Handler(
                version=0,
                permission=None,
                function=ping,
                name="ping",
                parameters={},
                documentation="Ping.",
                format=Format.PICKLE,
            ),
        )
        self.add_handler(
            0,
            "documentation",
            Handler(
                version=0,
                permission=None,
                function=documentation(self.name),
                name="documentation",
                parameters={},
                documentation="Documentation.",
                format=Format.PICKLE,
            ),
        )

    async def cleanup(self):
        await self.deferred_backend.cleanup_results()

    async def run(self):
        assert self.server is None, "Server already started."

        try:
            connection_parameters = {
                "host": self.host,
                "port": self.port,
            }

            if self.tls:
                connection_parameters["ssl"] = create_tls_context(server=True, configuration=self.tls)

            # fmt: off
            self.server = await asyncio.start_server(
                self._receive_wrapper,
                **connection_parameters,
            )
            # fmt: on
        except Exception as exception:
            context = {
                "exception": exception,
            }
            logger.exception(f"[Server] Exception while running server. Context `{context}`.")

    async def _receive_wrapper(self, reader, writer):
        def get_time():
            return time.time_ns() / 1000000000

        try:
            time_start = get_time()
            return await self._receive(ReaderWrapper(reader), WriterWrapper(writer))
        except asyncio.TimeoutError as exception:
            logger.error(f"[Server] Exception while serving request. Processing time is out.")
        except ConnectionError as exception:
            logger.error(f"[Server] Exception while serving request. Connection error.")
        except Exception as exception:
            context = {
                "exception": exception,
            }
            logger.exception(f"[Server] Exception while serving request. Context `{context}`.")

        try:
            writer.close()
        except:
            pass

        context = {
            "time": "{:0.6f}".format(get_time() - time_start),
        }
        logger.info(f"[Server] Function execution profiling. Context `{context}``.")

    async def _receive(self, reader, writer):
        # Read request.
        data_length_max = defaultdict(lambda: defaultdict(int))

        for version, handlers in self.handlers.items():
            for name, handler in handlers.items():
                if handler.data_length_max is None:
                    data_length_max[version][name] = self.name_length_max
                    continue

                data_length_max[version][name] = handler.data_length_max

        request = await Request.read(
            reader,
            token_length_max=self.token_length_max,
            name_length_max=self.name_length_max,
            data_length_max=data_length_max,
        )

        # Call handler.
        response = await self._process_request(request)

        # Write response.
        await response.write(writer)

        # Close connection.
        writer.close()

    async def _process_request(self, request):
        if request.version not in self.handlers:
            return Response(status=Status.FUNCTION_NOT_FOUND)

        if request.name not in self.handlers[request.version]:
            return Response(status=Status.FUNCTION_NOT_FOUND)

        handler = self.handlers[request.version][request.name]

        if request.documentation:
            return Response(status=Status.SUCCESS, documentation=handler.documentation)

        timeout = None

        if self.timeout:
            timeout = self.timeout

        if handler.timeout:
            timeout = handler.timeout

        async with asynchronous_context():
            if timeout:
                return await asyncio.wait_for(self._call_handler(request, handler), timeout=timeout)

            return await self._call_handler(request, handler)

    async def _call_handler(self, request, handler):
        parameters = {}

        # Check permission.
        if handler.permission:
            if not request.token:
                return Response(status=Status.AUTHENTICATION_MISSING_TOKEN)

            user = await self.authentication_backend.get_user(request.token)
            if not user:
                return Response(status=Status.AUTHENTICATION_USER_NOT_FOUND)

            has_permission = await user.has_permission(handler.permission)
            if not has_permission:
                return Response(status=Status.AUTHENTICATION_PERMISSION_DENIED)

            has_capacity = await user.has_capacity(handler.version, handler.name)
            if not has_capacity:
                return Response(status=Status.AUTHENTICATION_OUT_OF_CAPACITY)

            set_token(request.token)
            set_user(user)

        # Prevent deferred calls if not allowed by server or handler.
        if request.deferred and not self.deferred and not handler.allow_deferred:
            request.deferred = False

        # Prepare parameters.
        if request.data:
            if isinstance(request.data, dict):
                for key, value in request.data.items():
                    parameters[key] = value
            else:
                parameters["data"] = request.data

            # Unmarshall.
            # In case deferred request with operation id skip marshalling.
            if handler.schema_in and (not request.deferred or "operation_id" not in parameters):
                try:
                    parameters = handler.schema_in.load(parameters)
                except marshmallow.exceptions.ValidationError as exception:
                    return Response(status=Status.MARSHALLING_IN_FAILED)

        if handler.permission:
            parameters["user"] = user

        # TODO:
        # # Validate function signature.
        # if set(handler.parameters.keys()) != set(parameters.keys()):
        #     return Response(status=Status.FUNCTION_INVALID_SIGNATURE)

        # Call handler.
        async def _call_handler(operation_id=None):
            async def _wrapper():
                try:
                    result = await handler.function(**parameters)
                except:
                    logger.exception(f"[Server] Exception while handling request.")
                    return Response(status=Status.FAIL)

                # Marshall.
                if handler.schema_out:
                    try:
                        result = handler.schema_out.dump(result)
                    except marshmallow.exceptions.ValidationError as exception:
                        return Response(status=Status.MARSHALLING_OUT_FAILED)

                return result

            result = await _wrapper()

            # Return result immediately if not deferred operation.
            if operation_id is None:
                return result

            # Store result in case deferred operation.
            await self.deferred_backend.set_result(operation_id, result=result, done=True)

        # Decide what to return.
        if not request.deferred:
            result = await _call_handler()

            if isinstance(result, Response):
                return Response
        else:
            if "operation_id" in parameters:
                operation_id = parameters["operation_id"]
                result = await self.deferred_backend.get_result(operation_id)
            else:
                operation_id = self.deferred_backend.generate_operation_id()
                await self.deferred_backend.set_result(operation_id)
                asyncio.create_task(_call_handler(operation_id))
                result = operation_id

        # Select response format.
        response = Response(status=Status.SUCCESS, format=Format.PICKLE)

        if self.response_format:
            response.format = self.response_format

        if handler.format:
            response.format = handler.format

        # Cast handler result to response format.
        if isinstance(result, tuple):
            if len(result) == 2:
                response.status = result[0]
                response.data = result[1]

            if len(result) == 3:
                response.status = result[0]
                response.format = result[1]
                response.data = result[2]
        else:
            response.data = result

        # Return response.
        return response


async def ping():
    return


def documentation(server):
    async def _wrapper():
        lines = []

        for version, handlers in get_handlers().items():
            lines.append(f"Version {version}")
            lines.append("--------------------------------------------------------------------------------")

            for name, handler in handlers.items():
                if name in ("ping", "documentation"):
                    continue

                lines.append(handler.documentation)

        return "\n".join(lines)

    return _wrapper
